from __future__ import annotations

import tempfile
from pathlib import Path
from typing import List, Optional, Tuple

import nibabel as nib
import numpy as np
import torch
from loguru import logger
from nnunetv2.inference.predict_from_raw_data import nnUNetPredictor

from petu.constants import SEGMENTATION_LABELS, SEGMENTATION_THRESHOLD, InferenceMode
from petu.weights import check_weights_path


class ModelHandler:
    """Class for model loading, inference and post processing"""

    def __init__(self, device: torch.device) -> "ModelHandler":
        """Initialize the ModelHandler class.

        Args:
            device (torch.device): Device to use for inference.

        Returns:
            ModelHandler: ModelHandler instance.
        """

        self.device = device
        # Will be set during infer() call
        self.predictor = None
        self.inference_mode = None

        # get location of model weights
        self.model_weights_folder = check_weights_path()

    def load_model(self, inference_mode: InferenceMode) -> None:
        """Load the model for inference based on the inference mode

        Args:
            inference_mode (InferenceMode): inference mode (determined by passed images)
        """

        if not self.predictor or self.inference_mode != inference_mode:
            logger.debug(
                f"No loaded compatible model found (Switching from {self.inference_mode} to {inference_mode}). Loading Model and weights..."
            )
            self.inference_mode = inference_mode
            self.predictor = nnUNetPredictor(
                device=self.device,
            )
            self.predictor.initialize_from_trained_model_folder(
                self.model_weights_folder / self.inference_mode.value.replace("-", "_"),
                use_folds=("all"),
            )

            logger.debug(f"Successfully loaded model.")
        else:
            logger.debug(
                f"Same inference mode ({self.inference_mode}) as previous infer call. Re-using loaded model"
            )

    def threshold_probabilities(
        self, results_dir: Path
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Threshold the probabilities to get binary segmentations.

        Args:
            results_dir (Path): output directory of the nnUNet inference.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: Tuple of binary segmentations (ET, CC, T2H) and affine matrix.
        """

        nifti_file = next(results_dir.glob("*.nii.gz"))
        affine = nib.load(nifti_file).affine

        probabilities_file = next(results_dir.glob("*.npz"))
        probabilities = np.load(probabilities_file)["probabilities"]

        binary_data_list = []
        for i, _ in enumerate(SEGMENTATION_LABELS):
            transposed_data = np.transpose(probabilities[i], (2, 1, 0))
            binary_data = np.where(
                transposed_data > SEGMENTATION_THRESHOLD, 1, 0
            ).astype(np.int8)
            binary_data_list.append(binary_data)
        return *binary_data_list, affine

    def infer(
        self,
        input_file_paths: List[Path],
        ET_segmentation_file: Optional[str | Path] = None,
        CC_segmentation_file: Optional[str | Path] = None,
        T2H_segmentation_file: Optional[str | Path] = None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Run inference on the provided images and save the segmentations to disk if paths are provided.

        Args:
            input_file_paths (List[Path]): _description_
            ET_segmentation_file (Optional[str  |  Path], optional): Path to save ET segmentation. Defaults to None.
            CC_segmentation_file (Optional[str  |  Path], optional): Path to save CC segmentation. Defaults to None.
            T2H_segmentation_file (Optional[str  |  Path], optional): Path to save T2H segmentation. Defaults to None.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray]: Tuple of segmentations: (ET, CC, T2H) as numpy arrays.
        """

        with tempfile.TemporaryDirectory() as tmpdir:
            str_paths = [str(f) for f in input_file_paths]
            self.predictor.predict_from_files(
                [str_paths],
                tmpdir,
                save_probabilities=True,
            )
            et, cc, t2h, affine = self.threshold_probabilities(
                results_dir=Path(tmpdir),
            )

            # save segmentations to disk
            for data, path in zip(
                [et, cc, t2h],
                [ET_segmentation_file, CC_segmentation_file, T2H_segmentation_file],
            ):
                if path is not None:
                    path = Path(path)
                    path.parent.mkdir(parents=True, exist_ok=True)
                    nib.save(
                        nib.Nifti1Image(data, affine),
                        path,
                    )
                    logger.debug(f"Saved segmentation to {path}")

            return et, cc, t2h



import os
from petu.constants import NNUNET_ENV_VARS

# suppress nnUNet warnings by setting vars to dummy values (if not already set)
for env_var in NNUNET_ENV_VARS:
    if env_var not in os.environ:
        os.environ[env_var] = ""


from petu.inferer import Inferer



from __future__ import annotations

import os
import tempfile
from pathlib import Path
from typing import Dict, Optional, Tuple

import numpy as np
import torch
from loguru import logger

from petu.data_handler import DataHandler
from petu.model_handler import ModelHandler


class Inferer:

    def __init__(
        self,
        device: Optional[str] = "cuda",
        cuda_visible_devices: Optional[str] = "0",
    ) -> None:
        """
        Initialize the Inferer class.

        Args:
            device (Optional[str], optional): torch device string. Defaults to "cuda".
            cuda_visible_devices (Optional[str], optional): CUDA_VISIBLE_DEVICES environment variable. Defaults to "0".
        """
        self.device = self._configure_device(
            requested_device=device,
            cuda_visible_devices=cuda_visible_devices,
        )
        self.data_handler = DataHandler()
        self.model_handler = ModelHandler(device=self.device)

    def _configure_device(
        self, requested_device: str, cuda_visible_devices: str
    ) -> torch.device:
        """Configure the device for inference based on the specified config.device.

        Args:
            requested_device (str): Requested device.
            cuda_visible_devices (str): CUDA_VISIBLE_DEVICES environment variable.

        Returns:
            torch.device: Configured device.
        """
        device = torch.device(requested_device)
        if device.type == "cuda":
            # The env vars have to be set before the first call to torch.cuda, else torch will always attempt to use the first device
            os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
            os.environ["CUDA_VISIBLE_DEVICES"] = cuda_visible_devices
            if torch.cuda.is_available():
                # clean memory
                torch.cuda.empty_cache()

        logger.info(f"Set torch device: {device}")

        return device

    # @citation_reminder
    def infer(
        self,
        t1c: Optional[str | Path | np.ndarray] = None,
        fla: Optional[str | Path | np.ndarray] = None,
        t1: Optional[str | Path | np.ndarray] = None,
        t2: Optional[str | Path | np.ndarray] = None,
        ET_segmentation_file: Optional[str | Path] = None,
        CC_segmentation_file: Optional[str | Path] = None,
        T2H_segmentation_file: Optional[str | Path] = None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Infer segmentations based on provided images.

        Args:
            t1c (Optional[str  |  Path  |  np.ndarray], optional): T1C image. Defaults to None.
            fla (Optional[str  |  Path  |  np.ndarray], optional): FLAIR image. Defaults to None.
            t1 (Optional[str  |  Path  |  np.ndarray], optional): T1 image. Defaults to None.
            t2 (Optional[str  |  Path  |  np.ndarray], optional): T2 image. Defaults to None.
            ET_segmentation_file (Optional[str  |  Path], optional): Path to save ET segmentation. Defaults to None.
            CC_segmentation_file (Optional[str  |  Path], optional): Path to save CC segmentation. Defaults to None.
            T2H_segmentation_file (Optional[str  |  Path], optional): Path to save T2H segmentation. Defaults to None.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray]: Tuple of segmentations: (ET, CC, T2H) as numpy arrays.
        """

        # check inputs and get mode , if mode == prev mode => run inference, else load new model
        validated_images = self.data_handler.validate_images(
            t1=t1, t1c=t1c, t2=t2, fla=fla
        )
        determined_inference_mode = self.data_handler.determine_inference_mode(
            images=validated_images
        )

        self.model_handler.load_model(
            inference_mode=determined_inference_mode,
        )

        with tempfile.TemporaryDirectory() as tmpdir:

            input_file_paths = self.data_handler.get_input_file_paths(
                images=validated_images,
                tmp_folder=Path(tmpdir),
            )

            logger.info(f"Running inference on device: {self.device}")
            np_results = self.model_handler.infer(
                input_file_paths=input_file_paths,
                ET_segmentation_file=ET_segmentation_file,
                CC_segmentation_file=CC_segmentation_file,
                T2H_segmentation_file=T2H_segmentation_file,
            )
            logger.info(f"Finished inference")
            return np_results



from __future__ import annotations

import os
from pathlib import Path
from typing import List, Optional

import nibabel as nib
import numpy as np
from loguru import logger

from petu.constants import ATLAS_SPACE_SHAPE, IMGS_TO_MODE_DICT, DataMode, InferenceMode


class DataHandler:
    """Class to perform data related tasks such as validation, loading, transformation, saving."""

    def __init__(self) -> "DataHandler":
        # Following will be inferred from the input images during validation
        self.input_mode = None
        self.reference_nifti_file = None

    @property
    def get_input_mode(self) -> DataMode:
        """Get the input mode.

        Returns:
            DataMode: Input mode.
        Raises:
            AssertionError: If the input mode is not set (i.e. input images were not validated)
        """
        assert (
            self.input_mode is not None
        ), "Input mode not set. Please validate the input images first by calling .validate_images(...)."
        return self.input_mode

    @property
    def get_reference_nifti_file(self) -> Path | str:
        """Get a reference NIfTI file from the input (first not None in order T1-T1C-T2-FLAIR) to match header and affine.

        Returns:
            Path: Path to reference NIfTI file.
        Raises:
            AssertionError: If the reference NIfTI file is not set (i.e. input images were not validated)
        """
        assert (
            self.reference_nifti_file is not None
        ), "Reference NIfTI file not set. Please ensure you provided paths to NIfTI images and validated the input images first by calling .validate_images(...)."
        return self.reference_nifti_file

    def _validate_shape(self, data: np.ndarray) -> None:
        """Validate the shape of the input data.

        Args:
            data (np.ndarray): Input data.
        Raises:
            AssertionError: If the shape of the input data does not match the standard atlas space shape.
        """
        assert (
            data.shape == ATLAS_SPACE_SHAPE
        ), f"Invalid shape for input data, should match standard atlas space shape: {ATLAS_SPACE_SHAPE}"

    def validate_images(
        self,
        t1c: str | Path | np.ndarray | None = None,
        fla: str | Path | np.ndarray | None = None,
        t1: str | Path | np.ndarray | None = None,
        t2: str | Path | np.ndarray | None = None,
    ) -> List[np.ndarray | None] | List[Path | None]:
        """Validate inputs. \n
        Verify that the input images
            - exist (for paths)
            - are all of the same type (NumPy or NIfTI).
            - have the correct shape (ATLAS_SPACE_SHAPE).

        Sets internal variables input_mode and reference_nifti_file.

        Args:
            t1c (str | Path | np.ndarray | None, optional): T1C modality. Defaults to None.
            fla (str | Path | np.ndarray | None, optional): FLAIR modality. Defaults to None.
            t1 (str | Path | np.ndarray | None, optional): T1 modality. Defaults to None.
            t2 (str | Path | np.ndarray | None, optional): T2 modality. Defaults to None.

        Returns:
            List[np.ndarray | None] | List[Path | None]: List of validated images.
        Raises:
            FileNotFoundError: If a file is not found.
            ValueError: If a file path is not a NIfTI file (.nii or .nii.gz).
        """

        def _validate_image(
            data: str | Path | np.ndarray | None,
        ) -> np.ndarray | Path | None:
            if data is None:
                return None
            if isinstance(data, np.ndarray):
                self.input_mode = DataMode.NUMPY
                self._validate_shape(data)
                return data.astype(np.float32)

            data = Path(data)
            if not data.exists():
                raise FileNotFoundError(f"File {data} not found")

            if not (data.suffix == ".nii" or data.suffixes == [".nii", ".gz"]):
                raise ValueError(
                    f"File {data} must be a NIfTI file with extension .nii or .nii.gz"
                )

            self._validate_shape(nib.load(data).get_fdata())

            self.input_mode = DataMode.NIFTI_FILE
            return data.absolute()

        images = [
            _validate_image(img)
            for img in [
                t1c,
                fla,
                t1,
                t2,
            ]
        ]
        not_none_images = [img for img in images if img is not None]
        assert len(not_none_images) > 0, "No input images provided"
        # make sure all inputs have the same type
        unique_types = set(map(type, not_none_images))
        assert (
            len(unique_types) == 1
        ), f"All passed images must be of the same type! Received {unique_types}. Accepted Input types: {list(DataMode)}"

        if self.input_mode is DataMode.NIFTI_FILE:
            self.reference_nifti_file = not_none_images[0]
        logger.info(
            f"Successfully validated input images (received {len(not_none_images)}). Input mode: {self.input_mode}"
        )
        return images

    def determine_inference_mode(
        self, images: List[np.ndarray | None] | List[Path | None]
    ) -> InferenceMode:
        """Determine the inference mode based on the provided images.
        Args:
            images (List[np.ndarray | None] | List[Path | None]): List of validated images.
        Returns:
            InferenceMode: Inference mode based on the combination of input images.
        Raises:
            NotImplementedError: If no model is implemented for the combination of input images.
            AssertionError: If the input mode is not set (i.e. input images were not validated)
        """
        assert (
            self.input_mode is not None
        ), "Please validate the input images first by calling validate_images(...)."

        _t1c, _flair, _t1, _t2 = [img is not None for img in images]
        logger.debug(
            f"Received files: T1C: {_t1c}, FLAIR: {_flair}, T1: {_t1}, T2: {_t2}"
        )
        # check if files are given in a valid combination that has an existing model implementation
        mode = IMGS_TO_MODE_DICT.get((_t1c, _flair, _t1, _t2), None)
        if mode is None:
            raise NotImplementedError(
                f"No model implemented for this combination of images: T1C: {_t1c}, FLAIR: {_flair}, T1: {_t1}, T2: {_t2}. {os.linesep}Available models: {[mode.value for mode in InferenceMode]}"
            )
        logger.info(f"Inference mode: {mode}")
        return mode

    def get_input_file_paths(
        self,
        images: List[np.ndarray | None] | List[Path | None],
        tmp_folder: Optional[Path] = None,
    ) -> List[np.ndarray]:
        """Load the input images based on the input mode.

        Args:
            images (List[np.ndarray | None] | List[Path | None]): List of validated images.
        Returns:
            List[Path]: List of input images (NIFTI paths).
        Raises:
            AssertionError: If the input mode is not set (i.e. input images were not validated)
        """
        assert (
            self.input_mode is not None
        ), "Please validate the input images first by calling validate_images(...)."

        if self.input_mode == DataMode.NIFTI_FILE:
            return [img for img in images if img is not None]
        elif self.input_mode == DataMode.NUMPY:
            if tmp_folder is None:
                raise ValueError(
                    "Please provide a temporary folder when using NumPy input mode."
                )
            input_files = []
            for idx, img in enumerate(images):
                if img is not None:
                    input_file = tmp_folder / f"input_{idx}.nii.gz"
                    nib.save(nib.Nifti1Image(img, np.eye(4)), input_file)
                    input_files.append(input_file)
            return input_files
        else:
            raise NotImplementedError(
                f"Input mode {self.input_mode} not implemented. Available modes: {list(DataMode)}"
            )
